"""Main Archer API client."""

import logging
from typing import Optional
from dataclasses import dataclass

from .auth.authenticator import Authenticator
from .rest.applications import ApplicationsAPI
from .rest.questionnaires import QuestionnairesAPI
from .rest.records import RecordsAPI
from .rest.attachment import AttachmentAPI
from .rest.users import UsersAPI
from .rest.value_lists import ValueListsAPI
from .rest.reports import ReportsAPI
from .rest.search import SearchAPI
from .webservices.client import WebServicesClient
from .exceptions import ArcherAPIException


logger = logging.getLogger(__name__)


@dataclass
class ArcherConfig:
    """Configuration for Archer API client."""
    
    base_url: str
    instance_name: str
    timeout: int = 30
    max_retries: int = 3
    verify_ssl: bool = True
    proxy: Optional[dict] = None
    log_level: str = "INFO"


class RESTAPIClient:
    """Container for REST API resources."""
    
    def __init__(self, authenticator: Authenticator, config: ArcherConfig):
        self.applications = ApplicationsAPI(authenticator, config)
        self.questionnaires = QuestionnairesAPI(authenticator, config)
        self.records = RecordsAPI(authenticator, config)
        self.attachment = AttachmentAPI(authenticator, config)
        self.users = UsersAPI(authenticator, config)
        self.value_lists = ValueListsAPI(authenticator, config)
        self.reports = ReportsAPI(authenticator, config)
        self.search = SearchAPI(authenticator, config)


class ArcherClient:
    """
    Main client for interacting with Archer APIs.
    
    Args:
        base_url: Base URL of the Archer instance
        instance_name: Name of the Archer instance
        username: Optional username for automatic authentication
        password: Optional password for automatic authentication
        config: Optional ArcherConfig object
    """
    
    def __init__(
        self,
        base_url: str = None,
        instance_name: str = None,
        username: str = None,
        password: str = None,
        config: Optional[ArcherConfig] = None
    ):
        if config:
            self.config = config
        elif base_url and instance_name:
            self.config = ArcherConfig(
                base_url=base_url,
                instance_name=instance_name
            )
        else:
            raise ValueError("Either config or base_url and instance_name must be provided")
        
        logging.basicConfig(level=self.config.log_level)
        
        self._authenticator = Authenticator(self.config)
        self.rest = RESTAPIClient(self._authenticator, self.config)
        self.webservices = WebServicesClient(self._authenticator, self.config)
        
        if username and password:
            self.authenticate(username, password)
    
    def authenticate(self, username: str, password: str, user_domain: str = "") -> str:
        """Authenticate with Archer."""
        logger.info(f"Authenticating user: {username}")
        token = self._authenticator.authenticate(username, password, user_domain)
        logger.info("Authentication successful")
        return token
    
    def logout(self) -> None:
        """Logout and clear the session token."""
        logger.info("Logging out")
        self._authenticator.logout()
    
    @property
    def is_authenticated(self) -> bool:
        """Check if the client is authenticated."""
        return self._authenticator.is_authenticated
    
    @property
    def session_token(self) -> Optional[str]:
        """Get the current session token."""
        return self._authenticator.session_token


"""Record operations for Archer REST API with Pydantic model support."""

import logging
from typing import Dict, List, Optional, Any, Union

from ..exceptions import ResourceNotFoundError, ValidationError, ArcherAPIException
from ..models import (
    RecordCreateRequest,
    RecordUpdateRequest,
    RecordResponse,
    RecordContent,
    FieldContent
)
from .base import BaseAPI


logger = logging.getLogger(__name__)


class RecordsAPI(BaseAPI):
    """API client for Archer record operations."""
    
    def get(self, content_id: int, as_model: bool = False) -> Union[Dict[str, Any], RecordContent]:
        """
        Get a specific record by content ID.
        
        Args:
            content_id: The content ID of the record
            as_model: If True, return as RecordContent model, otherwise dict
            
        Returns:
            Record data as dictionary or RecordContent model
            
        Example:
            >>> # As dictionary (backward compatible)
            >>> record_dict = client.rest.records.get(123)
            
            >>> # As Pydantic model (with validation)
            >>> record_model = client.rest.records.get(123, as_model=True)
            >>> print(record_model.LevelId)
            >>> print(record_model.FieldContents)
        """
        url = f"{self.base_url}/api/core/content/{content_id}"
        
        try:
            response = self._make_request("GET", url)
            data = response.json()
            
            if as_model:
                # Parse and validate as Pydantic model
                return RecordContent(**data)
            return data
            
        except ArcherAPIException as e:
            if "404" in str(e):
                raise ResourceNotFoundError(f"Record {content_id} not found")
            raise
    
    def create(
        self,
        application_id: int = None,
        field_values: Dict[str, Any] = None,
        sub_form_data: Optional[List[Dict]] = None,
        request_model: Optional[RecordCreateRequest] = None
    ) -> int:
        """
        Create a new record in an application.
        
        You can use either the traditional parameters OR a RecordCreateRequest model.
        
        Args:
            application_id: ID of the application (traditional way)
            field_values: Dictionary mapping field names to values (traditional way)
            sub_form_data: Optional list of sub-form records (traditional way)
            request_model: RecordCreateRequest model (model-based way)
            
        Returns:
            Content ID of the created record
            
        Example (Traditional):
            >>> record_id = client.rest.records.create(
            ...     application_id=75,
            ...     field_values={
            ...         "Title": "New Incident",
            ...         "Priority": 1,
            ...         "Status": "Open"
            ...     }
            ... )
        
        Example (Model-based):
            >>> from py_ullr.models import RecordCreateRequest
            >>> 
            >>> request = RecordCreateRequest.from_fields(
            ...     application_id=75,
            ...     field_values={
            ...         "Title": "New Incident",
            ...         "Priority": 1,
            ...         "Status": "Open"
            ...     }
            ... )
            >>> record_id = client.rest.records.create(request_model=request)
        """
        url = f"{self.base_url}/api/core/content"
        
        # Use model if provided, otherwise build from parameters
        if request_model:
            payload = request_model.model_dump(exclude_none=True)
        else:
            if not application_id or not field_values:
                raise ValueError("Either request_model or (application_id and field_values) must be provided")
            
            # Build field content (traditional way)
            field_contents = {}
            for field_name, value in field_values.items():
                field_contents[field_name] = {
                    "Type": self._infer_field_type(value),
                    "Value": value
                }
            
            payload = {
                "Content": {
                    "LevelId": application_id,
                    "FieldContents": field_contents
                }
            }
            
            if sub_form_data:
                payload["Content"]["SubformData"] = sub_form_data
        
        try:
            response = self._make_request("POST", url, json=payload)
            data = response.json()
            
            # Optionally parse response as model
            response_model = RecordResponse(**data)
            
            if not response_model.IsSuccessful:
                raise ValidationError(f"Record creation failed: {response_model.ValidationMessages}")
            
            return response_model.record_id
            
        except ArcherAPIException as e:
            if "validation" in str(e).lower():
                raise ValidationError(f"Invalid field values: {e}")
            raise
    
    def update(
        self,
        content_id: int = None,
        field_values: Dict[str, Any] = None,
        request_model: Optional[RecordUpdateRequest] = None
    ) -> bool:
        """
        Update an existing record.
        
        Args:
            content_id: Content ID of the record to update (traditional way)
            field_values: Dictionary of field names and new values (traditional way)
            request_model: RecordUpdateRequest model (model-based way)
            
        Returns:
            True if update successful
            
        Example (Traditional):
            >>> success = client.rest.records.update(
            ...     content_id=123,
            ...     field_values={"Status": "Closed"}
            ... )
        
        Example (Model-based):
            >>> from py_ullr.models import RecordUpdateRequest, RecordContent, FieldContent
            >>> 
            >>> # First, get the existing record
            >>> existing = client.rest.records.get(123, as_model=True)
            >>> 
            >>> # Update specific fields
            >>> existing.FieldContents["Status"] = FieldContent(Type=1, Value="Closed")
            >>> 
            >>> request = RecordUpdateRequest(Content=existing)
            >>> success = client.rest.records.update(request_model=request)
        """
        url = f"{self.base_url}/api/core/content/{content_id or request_model.Content.Id}"
        
        if request_model:
            payload = request_model.model_dump(exclude_none=True)
        else:
            if not content_id or not field_values:
                raise ValueError("Either request_model or (content_id and field_values) must be provided")
            
            field_contents = {}
            for field_name, value in field_values.items():
                field_contents[field_name] = {
                    "Type": self._infer_field_type(value),
                    "Value": value
                }
            
            payload = {
                "Content": {
                    "Id": content_id,
                    "FieldContents": field_contents
                }
            }
        
        response = self._make_request("PUT", url, json=payload)
        return response.status_code == 200
    
    def delete(self, content_id: int) -> bool:
        """Delete a record."""
        url = f"{self.base_url}/api/core/content/{content_id}"
        response = self._make_request("DELETE", url)
        return response.status_code == 200
    
    def search(
        self,
        application_id: int,
        filters: Optional[Dict[str, Any]] = None,
        max_results: int = 100
    ) -> List[Dict[str, Any]]:
        """
        Search for records in an application (deprecated - use filter_by_reference).
        
        Note: This method uses a placeholder endpoint that may not exist in all Archer versions.
        """
        logger.warning("records.search() may not be supported. Consider using filter_by_reference() instead.")
        url = f"{self.base_url}/api/core/content/search"
        
        payload = {
            "LevelId": application_id,
            "MaxResultCount": max_results
        }
        
        if filters:
            payload["Filters"] = filters
        
        response = self._make_request("POST", url, json=payload)
        data = response.json()
        return data.get("RequestedObject", {}).get("Records", [])
    
    @staticmethod
    def _infer_field_type(value: Any) -> int:
        """Infer Archer field type from Python value."""
        if isinstance(value, str):
            return 1  # Text
        elif isinstance(value, (int, float)):
            return 2  # Numeric
        elif isinstance(value, bool):
            return 3  # Boolean
        elif isinstance(value, list):
            return 4  # Values List
        else:
            return 1  # Default to text



import pytest

from py_ullr.rest.records import RecordsAPI
from py_ullr.exceptions import ArcherAPIException, ResourceNotFoundError, ValidationError
from py_ullr.client import ArcherConfig


class DummyAuth:
    def __init__(self):
        self.is_authenticated = True

    def get_headers(self):
        return {}

    def get_session(self):
        return self._session

    def set_session(self, session):
        self._session = session


class MockResponse:
    def __init__(self, status_code=200, json_data=None, headers=None):
        self.status_code = status_code
        self._json = json_data or {}
        self.headers = headers or {}

    def json(self):
        return self._json

    def raise_for_status(self):
        if self.status_code >= 400:
            raise Exception("http error")


class MockSession:
    def __init__(self, response):
        self._response = response
        self.request_calls = []

    def request(self, method, url, headers=None, timeout=None, **kwargs):
        self.request_calls.append((method, url, headers, kwargs))
        return self._response


def test_create_returns_id_and_payload_contains_fieldcontents():
    dummy_auth = DummyAuth()
    response = MockResponse(status_code=201, json_data={"RequestedObject": {"Id": 123}})
    mock_session = MockSession(response)
    dummy_auth.set_session(mock_session)

    config = ArcherConfig(base_url="https://example", instance_name="inst")
    api = RecordsAPI(dummy_auth, config)

    record_id = api.create(10, {"Title": "x", "Count": 5})
    assert record_id == 123

    # Inspect the last request made by the mock session
    method, url, headers, kwargs = mock_session.request_calls[-1]
    assert "json" in kwargs
    assert kwargs["json"]["Content"]["LevelId"] == 10
    assert "FieldContents" in kwargs["json"]["Content"]


def test_get_raises_resource_not_found():
    dummy_auth = DummyAuth()
    config = ArcherConfig(base_url="https://example", instance_name="inst")
    api = RecordsAPI(dummy_auth, config)

    def raise_404(method, url, **kwargs):
        raise ArcherAPIException("404 Not Found")

    api._make_request = raise_404

    with pytest.raises(ResourceNotFoundError):
        api.get(999)


def test_create_validation_error_is_translated_to_validationerror():
    dummy_auth = DummyAuth()
    config = ArcherConfig(base_url="https://example", instance_name="inst")
    api = RecordsAPI(dummy_auth, config)

    def raise_validation(method, url, **kwargs):
        raise ArcherAPIException("validation error: bad")

    api._make_request = raise_validation

    with pytest.raises(ValidationError):
        api.create(1, {"f": "v"})



"""Authentication management for Archer APIs."""

import requests
import logging
from typing import Optional
from datetime import datetime, timedelta

from ..exceptions import AuthenticationError


logger = logging.getLogger(__name__)


class Authenticator:
    """Handles authentication and session management."""
    
    def __init__(self, config):
        self.config = config
        self._session_token: Optional[str] = None
        self._token_expiry: Optional[datetime] = None
        self._session = requests.Session()
        
        if config.proxy:
            self._session.proxies.update(config.proxy)
        
        self._session.verify = config.verify_ssl
    
    def authenticate(
        self,
        username: str,
        password: str,
        user_domain: str = ""
    ) -> str:
        """Authenticate with Archer and obtain a session token."""
        url = f"{self.config.base_url}/api/core/security/login"
        
        payload = {
            "InstanceName": self.config.instance_name,
            "Username": username,
            "UserDomain": user_domain,
            "Password": password
        }
        
        try:
            response = self._session.post(
                url,
                json=payload,
                timeout=self.config.timeout,
                headers={"Content-Type": "application/json"}
            )
            
            if response.status_code == 200:
                data = response.json()
                self._session_token = data.get("RequestedObject", {}).get("SessionToken")
                
                if not self._session_token:
                    raise AuthenticationError("No session token in response")
                
                self._token_expiry = datetime.now() + timedelta(minutes=30)
                
                logger.info("Authentication successful")
                return self._session_token
            else:
                error_msg = f"Authentication failed: {response.status_code}"
                logger.error(error_msg)
                raise AuthenticationError(error_msg)
                
        except requests.RequestException as e:
            logger.error(f"Request failed: {e}")
            raise AuthenticationError(f"Authentication request failed: {e}")
    
    def logout(self) -> None:
        """Logout and clear the session token."""
        if not self._session_token:
            return
        
        url = f"{self.config.base_url}/api/core/security/logout"
        
        try:
            self._session.post(
                url,
                headers=self._get_headers(),
                timeout=self.config.timeout
            )
        except requests.RequestException as e:
            logger.warning(f"Logout request failed: {e}")
        finally:
            self._session_token = None
            self._token_expiry = None
    
    def _get_headers(self) -> dict:
        """Get request headers with session token."""
        if not self._session_token:
            raise AuthenticationError("Not authenticated")
        
        return {
            "Authorization": f"Archer session-id={self._session_token}",
            "Content-Type": "application/json"
        }
    
    @property
    def is_authenticated(self) -> bool:
        """Check if currently authenticated with a valid token."""
        if not self._session_token:
            return False
        
        if self._token_expiry and datetime.now() >= self._token_expiry:
            logger.warning("Session token expired")
            return False
        
        return True
    
    @property
    def session_token(self) -> Optional[str]:
        """Get the current session token."""
        return self._session_token
    
    def get_session(self) -> requests.Session:
        """Get the configured requests session."""
        return self._session
    
    def get_headers(self) -> dict:
        """Get authenticated request headers."""
        return self._get_headers()


"""User operations for Archer REST API."""

import logging
from typing import List, Dict, Any

from .base import BaseAPI


logger = logging.getLogger(__name__)


class UsersAPI(BaseAPI):
    """API client for user operations."""
    
    def get_all(self) -> List[Dict[str, Any]]:
        """Get all users."""
        url = f"{self.base_url}/api/core/system/user"
        response = self._make_request("GET", url)
        return response.json()
    
    def get(self, user_id: int) -> Dict[str, Any]:
        """Get a specific user."""
        url = f"{self.base_url}/api/core/system/user/{user_id}"
        response = self._make_request("GET", url)
        return response.json()
    
    def search(self, username: str) -> List[Dict[str, Any]]:
        """Search for users by username."""
        url = f"{self.base_url}/api/core/system/user/search"
        payload = {"Username": username}
        response = self._make_request("POST", url, json=payload)
        return response.json()


"""Models for attachment operations."""

from typing import Optional
from pydantic import BaseModel, Field, validator
import base64


class AttachmentWrapper(BaseModel):
    """Wrapper for attachment data."""
    
    AttachmentName: str = Field(..., description="Name of the attachment file")
    AttachmentBytes: str = Field(..., description="Base64 encoded file bytes")
    IsSensitive: bool = Field(False, description="Whether to encrypt the attachment")
    
    @validator('AttachmentBytes')
    def validate_base64(cls, v):
        """Validate that AttachmentBytes is valid base64."""
        try:
            base64.b64decode(v)
        except Exception:
            raise ValueError("AttachmentBytes must be valid base64 encoded string")
        return v
    
    @classmethod
    def from_file(cls, filepath: str, is_sensitive: bool = False) -> "AttachmentWrapper":
        """
        Create AttachmentWrapper from a file path.
        
        Args:
            filepath: Path to the file
            is_sensitive: Whether to mark as sensitive/encrypted
            
        Returns:
            AttachmentWrapper instance
            
        Example:
            >>> attachment = AttachmentWrapper.from_file("document.pdf", is_sensitive=True)
        """
        import os
        
        filename = os.path.basename(filepath)
        
        with open(filepath, "rb") as f:
            file_bytes = f.read()
            encoded_bytes = base64.b64encode(file_bytes).decode('utf-8')
        
        return cls(
            AttachmentName=filename,
            AttachmentBytes=encoded_bytes,
            IsSensitive=is_sensitive
        )
    
    def save_to_file(self, output_path: str) -> None:
        """
        Save attachment bytes to a file.
        
        Args:
            output_path: Path where to save the file
            
        Example:
            >>> attachment.save_to_file("downloaded_document.pdf")
        """
        decoded_bytes = base64.b64decode(self.AttachmentBytes)
        
        with open(output_path, "wb") as f:
            f.write(decoded_bytes)


class AttachmentUploadRequest(BaseModel):
    """Request for uploading an attachment."""
    
    ContentId: int = Field(..., description="Content record ID")
    FieldId: int = Field(..., description="Attachment field ID")
    AttachmentName: str = Field(..., description="File name")
    IsSensitive: bool = Field(False, description="Encrypt the file")


class AttachmentResponse(BaseModel):
    """Response from attachment operations."""
    
    Links: list = Field(default_factory=list)
    RequestedObject: dict = Field(default_factory=dict)
    IsSuccessful: bool
    ValidationMessages: list = Field(default_factory=list)
    
    @property
    def attachment_id(self) -> Optional[int]:
        """Extract attachment ID from response."""
        return self.RequestedObject.get("Id")


"""Models for Archer questionnaire operations."""

from typing import List, Optional, Dict, Any
from pydantic import BaseModel, Field


class QuestionnaireRule(BaseModel):
    """Represents a questionnaire rule in Archer."""
    
    Id: int = Field(..., description="Questionnaire rule ID")
    Name: Optional[str] = Field(None, description="Rule name")
    QuestionnaireId: Optional[int] = Field(None, description="Parent questionnaire ID")
    LevelId: Optional[int] = Field(None, description="Target application/level ID")
    
    class Config:
        json_schema_extra = {
            "example": {
                "Id": 501,
                "Name": "Auto-populate Rule",
                "QuestionnaireId": 100,
                "LevelId": 75
            }
        }


class Questionnaire(BaseModel):
    """Represents an Archer questionnaire."""
    
    Id: int = Field(..., description="Questionnaire ID")
    Name: str = Field(..., description="Questionnaire name")
    Alias: Optional[str] = Field(None, description="Questionnaire alias")
    Type: Optional[int] = Field(None, description="Questionnaire type")
    LanguageId: Optional[int] = Field(None, description="Language ID")
    
    # Target application reference
    TargetApplicationId: Optional[int] = Field(
        None,
        description="ID of the target application this questionnaire links to"
    )
    TargetApplicationName: Optional[str] = Field(
        None,
        description="Name of the target application"
    )
    
    # Optional metadata
    fields: Optional[List[Dict[str, Any]]] = Field(
        None,
        description="Questionnaire fields"
    )
    rules: Optional[List[QuestionnaireRule]] = Field(
        None,
        description="Questionnaire rules"
    )
    
    class Config:
        json_schema_extra = {
            "example": {
                "Id": 100,
                "Name": "Security Assessment Questionnaire",
                "Alias": "SEC_ASSESS_Q",
                "Type": 1,
                "LanguageId": 1,
                "TargetApplicationId": 75,
                "TargetApplicationName": "Security Incidents"
            }
        }
    
    def get_rule_by_id(self, rule_id: int) -> Optional[QuestionnaireRule]:
        """
        Get questionnaire rule by ID.
        
        Args:
            rule_id: ID of the rule
            
        Returns:
            QuestionnaireRule if found, None otherwise
        """
        if not self.rules:
            return None
        
        return next((r for r in self.rules if r.Id == rule_id), None)
    
    def get_rules_by_level(self, level_id: int) -> List[QuestionnaireRule]:
        """
        Get all rules for a specific level/application.
        
        Args:
            level_id: ID of the level/application
            
        Returns:
            List of QuestionnaireRule for the level
        """
        if not self.rules:
            return []
        
        return [r for r in self.rules if r.LevelId == level_id]



"""Models for search and filtering operations."""

from typing import Dict, List, Optional, Any
from enum import Enum
from pydantic import BaseModel, Field


class SearchOperator(str, Enum):
    """Search operators for filtering."""
    EQUALS = "eq"
    NOT_EQUALS = "ne"
    GREATER_THAN = "gt"
    GREATER_THAN_OR_EQUAL = "ge"
    LESS_THAN = "lt"
    LESS_THAN_OR_EQUAL = "le"
    CONTAINS = "contains"
    STARTS_WITH = "startswith"
    ENDS_WITH = "endswith"


class SearchFilter(BaseModel):
    """Represents a single search filter criterion."""
    
    field_name: str = Field(..., description="Name or ID of the field to filter")
    operator: SearchOperator = Field(..., description="Comparison operator")
    value: Any = Field(..., description="Value to compare against")
    
    def to_odata_string(self) -> str:
        """
        Convert filter to OData query string format.
        
        Returns:
            OData filter string
            
        Example:
            >>> filter = SearchFilter(field_name="Status", operator=SearchOperator.EQUALS, value="Active")
            >>> filter.to_odata_string()
            "Status eq 'Active'"
        """
        if isinstance(self.value, str):
            value_str = f"'{self.value}'"
        else:
            value_str = str(self.value)
        
        if self.operator in [SearchOperator.CONTAINS, SearchOperator.STARTS_WITH, SearchOperator.ENDS_WITH]:
            return f"{self.operator.value}({self.field_name}, {value_str})"
        else:
            return f"{self.field_name} {self.operator.value} {value_str}"
    
    class Config:
        json_schema_extra = {
            "example": {
                "field_name": "Status",
                "operator": "eq",
                "value": "Active"
            }
        }


class SearchCriteria(BaseModel):
    """Complete search criteria for querying records."""
    
    filters: List[SearchFilter] = Field(default_factory=list)
    select_fields: Optional[List[str]] = Field(None, description="Fields to return")
    order_by: Optional[str] = Field(None, description="Field to order by")
    top: Optional[int] = Field(None, ge=1, description="Number of records to return")
    skip: Optional[int] = Field(None, ge=0, description="Number of records to skip")
    
    def build_query_string(self) -> str:
        """
        Build OData query string from criteria.
        
        Returns:
            Query string parameters
            
        Example:
            >>> criteria = SearchCriteria(
            ...     filters=[SearchFilter(field_name="Status", operator=SearchOperator.EQUALS, value="Active")],
            ...     top=10
            ... )
            >>> criteria.build_query_string()
            "$filter=Status eq 'Active'&$top=10"
        """
        parts = []
        
        if self.filters:
            filter_str = " and ".join([f.to_odata_string() for f in self.filters])
            parts.append(f"$filter={filter_str}")
        
        if self.select_fields:
            parts.append(f"$select={','.join(self.select_fields)}")
        
        if self.order_by:
            parts.append(f"$orderby={self.order_by}")
        
        if self.top is not None:
            parts.append(f"$top={self.top}")
        
        if self.skip is not None:
            parts.append(f"$skip={self.skip}")
        
        return "&".join(parts)


class SearchResponse(BaseModel):
    """Response from search operations."""
    
    records: List[Dict[str, Any]] = Field(default_factory=list)
    total_count: Optional[int] = None
    
    @property
    def count(self) -> int:
        """Get the number of records returned."""
        return len(self.records)


"""REST API module."""

from .applications import ApplicationsAPI
from .questionnaires import QuestionnairesAPI
from .records import RecordsAPI
from .attachment import AttachmentAPI
from .users import UsersAPI
from .value_lists import ValueListsAPI
from .reports import ReportsAPI
from .search import SearchAPI

__all__ = [
    "ApplicationsAPI",
    "QuestionnairesAPI",
    "RecordsAPI",
    "AttachmentAPI",
    "UsersAPI",
    "ValueListsAPI",
    "ReportsAPI",
    "SearchAPI",
]


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="py-ullr",
    version="1.0.3",
    author="Takeshi Cho",
    author_email="chotakeshi.richard@gmail.com",
    description="A comprehensive Python library for Archer IRM API integration",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.8",
    install_requires=[
        "requests>=2.28.0",
        "requests-toolbelt>=0.10.0",
        "python-dateutil>=2.8.0",
        "pydantic>=2.0.0",
    ],
    extras_require={
        "dev": [
            "pytest>=7.0.0",
            "pytest-cov>=4.0.0",
            "black>=23.0.0",
            "flake8>=6.0.0",
            "mypy>=1.0.0",
        ],
    },
)


"""Application operations for Archer REST API with model support."""

import logging
from typing import List, Dict, Any, Union

from ..models.application import Application, FieldDefinition
from .base import BaseAPI


logger = logging.getLogger(__name__)


class ApplicationsAPI(BaseAPI):
    """API client for Archer application operations."""
    
    def get_all(self, as_model: bool = False) -> Union[List[Dict[str, Any]], List[Application]]:
        """Get all applications."""
        url = f"{self.base_url}/platformapi/core/system/application"
        response = self._make_request("GET", url)
        data = response.json()
        
        # Extract from RequestedObject if wrapped
        if isinstance(data, dict) and 'RequestedObject' in data:
            data = data['RequestedObject']
        
        if as_model:
            # Handle both list and single object
            if isinstance(data, list):
                return [Application(**app) for app in data]
            else:
                return [Application(**data)]
        return data
    
    def get(
        self,
        application_id: int,
        as_model: bool = False,
        include_fields: bool = False
    ) -> Union[Dict[str, Any], Application]:
        """
        Get a specific application.
        
        Args:
            application_id: Application ID
            as_model: If True, return as Application model
            include_fields: If True and as_model=True, fetch and include field definitions
            
        Returns:
            Application data as dict or Application model
            
        Example:
            >>> # Method 1: Get as dictionary (traditional)
            >>> app_data = client.rest.applications.get(application_id=191)
            >>> print(app_data)
            >>> 
            >>> # Method 2: Get as model with validation
            >>> app = client.rest.applications.get(application_id=191, as_model=True)
            >>> print(f"App: {app.Name} (ID: {app.Id})")
            >>> 
            >>> # Method 3: Get with fields included
            >>> app = client.rest.applications.get(
            ...     application_id=191,
            ...     as_model=True,
            ...     include_fields=True
            ... )
            >>> for field in app.fields:
            ...     print(f"{field.Name}: {field.type_name}")
        """
        url = f"{self.base_url}/platformapi/core/system/application/{application_id}"
        response = self._make_request("GET", url)
        data = response.json()
        
        # ✅ KEY FIX: Extract from RequestedObject wrapper
        if isinstance(data, dict) and 'RequestedObject' in data:
            data = data['RequestedObject']
        
        if as_model:
            app = Application(**data)
            
            # Optionally fetch and attach field definitions
            if include_fields:
                fields_data = self.get_fields(application_id)
                app.fields = [FieldDefinition(**f) for f in fields_data]
            
            return app
        
        return data
    
    def get_fields(
        self,
        application_id: int,
        as_model: bool = False
    ) -> Union[List[Dict[str, Any]], List[FieldDefinition]]:
        """
        Get fields for an application.
        
        Args:
            application_id: Application ID
            as_model: If True, return as list of FieldDefinition models
            
        Returns:
            List of field definitions
        """
        url = f"{self.base_url}/platformapi/core/system/fielddefinition/application/{application_id}"
        response = self._make_request("GET", url)
        data = response.json()
        
        # Extract from RequestedObject if wrapped
        if isinstance(data, dict) and 'RequestedObject' in data:
            data = data['RequestedObject']
        
        # Ensure data is a list
        if not isinstance(data, list):
            data = [data] if data else []
        
        if as_model:
            return [FieldDefinition(**f) for f in data]
        return data


"""
Archer API Client Library
~~~~~~~~~~~~~~~~~~~~~~~~~

A comprehensive Python library for Archer IRM API integration.
"""

__version__ = "1.0.0"
__author__ = "Takeshi Cho"

from .client import ArcherClient, ArcherConfig
from .exceptions import (
    ArcherAPIException,
    AuthenticationError,
    ResourceNotFoundError,
    ValidationError,
    RateLimitError
)

__all__ = [
    "ArcherClient",
    "ArcherConfig",
    "ArcherAPIException",
    "AuthenticationError",
    "ResourceNotFoundError",
    "ValidationError",
    "RateLimitError",
]


"""Data models for Archer API."""

# Add Pydantic models here as needed

from .record import (
    FieldContent,
    FieldType,
    RecordContent,
    RecordResponse,
    RecordCreateRequest,
    RecordUpdateRequest
)
from .search import (
    SearchOperator,
    SearchFilter,
    SearchCriteria,
    SearchResponse
)
from .attachment import (
    AttachmentWrapper,
    AttachmentUploadRequest,
    AttachmentResponse
)
from .application import (
    Application,
    FieldDefinition,
    FieldTypeEnum
)
from .questionnaire import (
    Questionnaire,
    QuestionnaireRule
)

__all__ = [
    # Record models
    "FieldContent",
    "FieldType",
    "RecordContent",
    "RecordResponse",
    "RecordCreateRequest",
    "RecordUpdateRequest",
    # Search models
    "SearchOperator",
    "SearchFilter",
    "SearchCriteria",
    "SearchResponse",
    # Attachment models
    "AttachmentWrapper",
    "AttachmentUploadRequest",
    "AttachmentResponse",
    # Application models
    "Application",
    "Questionnaire",
    "QuestionnaireRule",
    "FieldDefinition",
    "FieldTypeEnum",
]